#!/bin/sh

# This script runs orcallator.se with the proper options for your site.

# Define program locations that will be needed.
prefix=/usr/local/orca-0.27
exec_prefix=${prefix}
libdir=${exec_prefix}/lib
AWK=/usr/bin/nawk
COMPRESSOR="/usr/bin/bzip2 -9"
CUT=/usr/bin/cut
EXPR=/usr/bin/expr
UNAME=/usr/bin/uname
ORCALLATOR_DIR=/usr/local/orca-0.27/var/orca/orcallator
SE=/opt/RICHPse/bin/se

# WEB_LOG contains the location of the web server log file that
# orcallator.se should read.
WEB_LOG=

# WEB_SERVER contains a regular expression used to find the number of
# running non-secure web server processes on the system.
WEB_SERVER=httpd

# WEB_SERVER_SECURE contains a regular expression used to find the
# number of running secure web server processes on the system.
WEB_SERVER_SECURE=httpsd

# WATCH_WEB contains the command line options to SE to tell
# orcallator.se the format of the web server logs.
WATCH_WEB="-DWATCH_WEB"               # For NCSA style access log
WATCH_WEB="-DWATCH_WEB -DWATCH_PROXY" # For proxy NCSA style access log
WATCH_WEB="-DWATCH_WEB -DWATCH_SQUID" # For Squid style access log
WATCH_WEB=""

# These are the various patch defines to make sure SE works around
# various problems.
SE_PATCHES=
#SE_PATCHES="$SE_PATCHES -DLE_PATCH"
#SE_PATCHES="$SE_PATCHES -DHME_PATCH"
#SE_PATCHES="$SE_PATCHES -DHME_PATCH_IFSPEED"

# Check if the SE executable was found upon configure.
if test -z "$SE"; then
  cat <<END 1>&2
When you configured Orca the se executable was not found.  If you
do not have the SE toolkit installed on your system, then follow
the steps in section 10 of INSTALL file that came with Orca.

Once you have the SE toolkit installed on your system, then either
rerun configure so that it finds se, or edit start_orcallator.sh
and define SE to the location of se.
END
  exit 1
fi

if test ! -x "$SE"; then
  echo "$0: the SE executable at $SE does not exist or is not executable." 1>&2
  exit 1
fi

# Get the hostname without the fully qualified part; that is, trim off
# anything past the first `.'.
uname=`$UNAME -n | $CUT -d. -f1`

# Make sure the hostname can be found.
if test -z "$uname"; then
  echo "$0: cannot determine the hostname of this system." 1>&2
  exit 1
fi

# The directory these files go into is $ORCALLATOR_DIR/HOSTNAME
OUTDIR=$ORCALLATOR_DIR/$uname

# Export the environmental variables.
export COMPRESSOR OUTDIR WEB_LOG WEB_SERVER WEB_SERVER_SECURE

# Check if orcallator is already running.
pids=`/usr/ucb/ps auxww | $AWK '/orcallator.se/ && !/awk/ {print $2}'`
if test "$pids" != ""; then
  echo "$0: orcallator already running." 1>&2
  exit 1
fi

echo "Writing data into $OUTDIR/"
if test "$WEB_LOG"; then
  echo "Using www access log file $WEB_LOG"
fi

# Cd to / so that any automounted filesystems can be unmounted.
cd /

# Create the output directory if it doesn't exist yet.
if test ! -d $OUTDIR; then
  echo "Creating $OUTDIR/"
  mkdir -p $OUTDIR
fi

if test ! -d $OUTDIR; then
  echo "$0: unable to create $OUTDIR/" 1>&2
  exit 2
fi

# Now start the logging.
echo "Starting logging"
nohup $SE $SE_PATCHES -DWATCH_OS $WATCH_WEB $libdir/orcallator.se &

# Write the PID of orcallator to a file to make killing easier.
pid=$!
echo $pid > $OUTDIR/orcallator.pid

# Sleep for a couple of seconds to allow any orcallator warnings to appear.
sleep 5
